import SeeThru_Feeds.Model.Properties.Properties
from SeeThru_Feeds.Model.Properties.Properties import PropertyBase


class PropertyManager:
    def __init__(self):
        self.Properties = None
        self.initialise()

    def initialise(self):
        """
        Ensures that the Properties attribute is set
        """
        # Initialises the properties dictionary for the manager
        # This stores new instances of every property defined
        if self.Properties is None or type(self.Properties) is not dict:
            self.Properties = {}
            # --Fills the properties dict with the the subclass' properties
            # Loops through every variable in the class, getting the
            # variable name and it's associated value
            for attr in self.__dir__():
                prop = getattr(self, attr)
                # If the attribute is a property, then it is added to the properties dict
                if isinstance(prop, SeeThru_Feeds.Model.Properties.Properties.PropertyBase):
                    # Uses either a name defined in the property or the variable's name
                    name = prop.name if prop.name is not None else attr

                    # Creates a new instance of the property
                    # This is used to store the property value
                    instanceProp = prop.new_instance(prop)
                    instanceProp._value = instanceProp.default  # Assigns the default value, ! This bypasses validation
                    # Stores the instance property in the Properties dictionary
                    self.Properties[name] = instanceProp

                    # Sets the properties internal name to the name obtained
                    instanceProp.name = name

                    # Makes the property instance available 
                    # by both the variable name and internal name
                    self.__setattr__(attr, instanceProp)
                    self.__setattr__(name, instanceProp)

    def set_property(self, prop, value):
        """
        Sets a value to the given property
        Note: The value isn't stored in the property instance

        Arguments:
            prop (PropertyBase): The value to associate with the property
            value (any): The value to give the property

        Raises:
            TypeError: Prop argument must be either of type Property or string
            TypeError: Modifier argument must be a function pointer

        Returns:
            PropertyManager: The property manager
        """
        self.initialise()
        # Data validation
        if not isinstance(prop, SeeThru_Feeds.Model.Properties.Properties.PropertyBase) and type(prop) != str:
            raise TypeError("Prop argument must be either of type Property or string")
        # Gets the index name of the property
        # it is either the value passed in as a string or the prop's internal name
        indexName = prop if type(prop) == str else prop.name
        # Assigns the new value to the property
        self.Properties[indexName].value = value
        return self

    def get_property(self, prop):
        """
        Gets the value associated with a property in the manager

        Arguments:
            prop (PropertyBase, str): The property reference (e.g. HTTPGet.URL) or the property name

        Raises:
            TypeError: Prop argument must be either of type Property or string

        Returns:
            any: The value
        """
        self.initialise()
        # Data validation
        if not isinstance(prop, SeeThru_Feeds.Model.Properties.Properties.PropertyBase) and type(prop) != str:
            raise TypeError(
                "Prop argument must be either of type Property or string")
        # Gets the index name of the property
        # it is either the value passed in as a string or the prop's internal name
        indexName = prop if type(prop) == str else prop.name
        # Returns the value of the property
        return self.Properties[indexName].value

    def modify_property(self, prop, modifier):
        """
        Modifies the value of a given property

        Arguments:
            prop (PropertyBase, str): The property reference (e.g. HTTPGet.URL) or the property name
            modifier (lambda): A function which takes the value as an argument and modifies it in some way

        Raises:
            TypeError: Prop argument must be either of type Property or string
            TypeError: Modifier argument must be a function pointer

        Returns:
            PropertyManager: The property manager
        """
        self.initialise()
        # Data validation
        if not isinstance(prop, SeeThru_Feeds.Model.Properties.Properties.PropertyBase) and type(prop) != str:
            raise TypeError(
                "Prop argument must be either of type Property or string")
        if not callable(modifier):
            raise TypeError("modifier argument must be a function pointer")
        # Gets the index name of the property
        # it is either the value passed in as a string or the prop's internal name
        indexName = prop if type(prop) == str else prop.name
        # Assigns the new value to the property
        self.Properties[indexName].value = modifier(
            self.Properties[indexName].value)
        return self

    @property
    def FillableProperties(self):
        """Returns a list of the fillable properties for the manager

        Returns:
            list: A list of fillable result properties
        """
        self.initialise()
        # Loops through all of the properties and returns the names of the fillable
        fillable = [prop for prop in self.Properties.values() if isinstance(prop, SeeThru_Feeds.Model.Properties.Properties.FillableProperty)]
        return fillable

    @property
    def ResultProperties(self):
        """Returns a list of the result properties for the manager

        Returns:
            list: A list of result property instances
        """
        self.initialise()
        # Loops through all of the properties and returns the names of the result properties
        results = [prop for prop in self.Properties.values() if isinstance(prop, SeeThru_Feeds.Model.Properties.Properties.ResultProperty)]
        return results

    def check_fillables(self):
        """
        Parses all of the fillable properties to make sure that they are valid

        Raises:
            Exception: Did not pass a property pass

        Returns:
            bool: Whether the fillables are valid, effectively it is always true as exceptions are made
        """
        self.initialise()
        # Parses all of the fillable properties of the manager to make sure that they are met
        for prop in self.FillableProperties:
            if not prop.parse_value(prop.value):
                raise Exception("[Property: {}] Did not pass a property parse".format(prop.name))
        return True



import json

from SeeThru_Feeds.Model.Scripts.ScriptResult import ScriptResult

import requests


class Feed:
    def __init__(self):
        self.AccessToken = None
        self.Secret = None
        self.Guid = None
        self.ScriptResult = None

    def set_api_key(self, access_token, secret):
        self.AccessToken = access_token
        self.Secret = secret

    def set_script(self, script):
        self.ScriptResult = script.get_result()

    def set_guid(self, guid):
        self.Guid = guid

    def set_script_result(self, result):
        self.ScriptResult = result

    def push(self):
        headers = {
            "X-Access-Token": self.AccessToken,
            "X-Secret": self.Secret
        }
        data = {
            "guid": self.Guid,
            "color": self.ScriptResult.get_status(),
            "message": self.ScriptResult.get_message(),
            "timestamp": self.ScriptResult.get_timestamp(),
            "version": "2020.1.1"
        }

        response = requests.post(f"https://api.seethrunetworks.com/feed/{self.Guid}/update", data=data, headers=headers)

        if response.status_code != 200:
            return False
        response_json = json.loads(response.content)
        if not response_json["success"]:
            return response_json["message"]
        return ""



from SeeThru_Feeds.Model.Components.ComponentBase import ComponentBase
from SeeThru_Feeds.Model.Properties.Properties import FillableProperty, ResultProperty
import mysql.connector


class ConnectionOpen(ComponentBase):
    HOST = FillableProperty(name="Host", required=True)
    USER = FillableProperty(name="User", required=True)
    PASSWD = FillableProperty(name="password", required=True)
    DATABASE = FillableProperty(name="Database", required=True)
    CAN_CONNECT = ResultProperty(name="Can_Connect")

    Component_Title = "ConnectionOpen Database Component"
    Component_Description = "This component tests to see if it can open a database connection with the given details"
    Component_Author = "SeeThru Networks"
    Component_Owner = "SeeThru Networks"

    def component_execute(self):
        try:
            connection = mysql.connector.connect(
                host=self.get_property(ConnectionOpen.HOST),
                user=self.get_property(ConnectionOpen.USER),
                passwd=self.get_property(ConnectionOpen.PASSWD),
                database=self.get_property(ConnectionOpen.DATABASE)
            )
            connection.close()
            # Assumes that a valid connection was made
            self.set_property(ConnectionOpen.CAN_CONNECT, True)
        except:
            self.set_property(ConnectionOpen.CAN_CONNECT, False)



from typing import Dict

from . import Header as ConfigHeader, Script as ConfigScript, Feed as ConfigFeed, ApiKey as ConfigApiKey

from SeeThru_Feeds.Core.Exceptions import Config as ConfigExceptions

class Config:
    Header: "ConfigHeader.Header"
    Scripts: Dict[str, "ConfigScript.Script"]
    Feeds: Dict[str, "ConfigFeed"]
    Api_Keys: Dict[str, "ConfigApiKey.ApiKey"]

    def __init__(self,
                 header: "ConfigHeader.Header",
                 scripts: Dict[str, "ConfigScript.Script"],
                 feeds: Dict[str, "ConfigFeed.Feed"],
                 api_keys: Dict[str, "ConfigApiKey.ApiKey"]):
        self.Header = header
        self.Scripts = scripts
        self.Feeds = feeds
        self.Api_Keys = api_keys

    @staticmethod
    def new(name: str) -> "Config":
        """
        Creates a new config, with newly generated attributes

        Args:
            name (str): The name of the feedscheme

        Returns:
            Config: The new config
        """
        header = ConfigHeader.Header.new(name)
        return Config(header, {}, {}, {})

    def add_script(self, name: str) -> ConfigScript:
        """
        Adds a script to the config

        Args:
            name (str): The name of the script

        Returns:
            Script: The new script
        """
        configscript = ConfigScript.Script.new(name)
        self.Scripts[name] = configscript
        return configscript

    def add_feed(self, name: str, _script: str, api_key: str, guid: str) -> "ConfigFeed.Feed":
        """
        Adds a feed to the config

        Args:
            name (str): The name of the feed
            _script (str): The name of the script that the feed uses
            api_key (str): The name of the api key that the feed uses
            guid (str): The guid of the feed

        Returns:
            Feed: The new feed
        """
        feed = ConfigFeed.Feed.new(_script, api_key, guid)
        self.Feeds[name] = feed
        return feed

    def add_api_key(self, name: str, access_token: str, secret: str) -> ConfigApiKey.ApiKey:
        """
        Adds an api key to the config

        Args:
            name (str): The name of the api key
            access_token (str): The access token of the api key
            secret (str): The secret key of the api key

        Returns:
            ApiKey: The new api key
        """
        api_key = ConfigApiKey.ApiKey.new(access_token, secret)
        self.Api_Keys[name] = api_key
        return api_key

    @staticmethod
    def load(data: dict) -> "Config":
        if "Header" in data:
            header = ConfigHeader.Header.load(data["Header"])
        else:
            raise ConfigExceptions.ConfigException("No header provided")
        scripts = {}
        if "Scripts" in data:
            if type(data["Scripts"]) != dict:
                raise ConfigExceptions.ConfigException("Config has invalid scripts")
            scripts = {
                key: ConfigScript.Script.load(data["Scripts"][key], key) for key in data["Scripts"].keys()
            }
        feeds = {}
        if "Feeds" in data:
            if type(data["Feeds"]) != dict:
                raise ConfigExceptions.ConfigException("Config has invalid feeds")
            feeds = {
                key: ConfigFeed.Feed.load(data["Feeds"][key], key) for key in data["Feeds"].keys()
            }
        api_keys = {}
        if "Api_Keys" in data:
            if type(data["Api_Keys"]) != dict:
                raise ConfigExceptions.ConfigException("Config has invalid api_keys")
            api_keys = {
                key: ConfigApiKey.ApiKey.load(data["Api_Keys"][key], key) for key in data["Api_Keys"].keys()
            }
        return Config(
            header,
            scripts,
            feeds,
            api_keys
        )

    def dump(self) -> dict:
        data = {
            "Header": self.Header.dump()
        }
        if len(self.Scripts) != 0:
            data["Scripts"] = {
                key: self.Scripts[key].dump() for key in self.Scripts.keys()
            }
        if len(self.Feeds) != 0:
            data["Feeds"] = {
                key: self.Feeds[key].dump() for key in self.Feeds.keys()
            }
        if len(self.Api_Keys) != 0:
            data["Api_Keys"] = {
                key: self.Api_Keys[key].dump() for key in self.Api_Keys.keys()
            }
        return data



from SeeThru_Feeds.Model.Attribution import Attribution
from SeeThru_Feeds.Model.Scripts.ScriptResult import ScriptResult
from SeeThru_Feeds.Model.Properties.PropertyManager import PropertyManager
from SeeThru_Feeds.Model.Feeds.Feed import Feed
from pathlib import Path

import SeeThru_Feeds.Model.Scripts.ScriptState as ScriptState


class ScriptBase(PropertyManager, Attribution):
    def __init__(self, *args, **kwargs):
        """
        Initialises the script
        ! If you must use __init__ in the subclass, you must call super().__init__() from that initializer
        ! otherwise override prepare and that'll be called after the super class is initialised
        """
        super().__init__()
        self._script_alias = None
        self._script_result = None
        self._script_output_path = None
        self._guid = None

        self.prepare(*args, **kwargs)

    def prepare(self, *args, **kwargs):
        """
        This can be overridden if there *needs* to be special preparation steps
        """
        pass

    # region Run

    def script_run(self):
        """
        This function should be overridden by a subclass
        This is where your script should start running

        Raises:
            NotImplementedError: There is no execution method defined, please define it with 'script_run'
        """
        raise NotImplementedError("There is no execution method defined, please define it with 'script_run'")

    def run_script(self):
        """
        Runs the script

        Returns:
            ScriptBase: The script
        """
        try:
            # All fillables are checked on the setter, this performs one final check
            self.check_fillables()
            self.script_run()
        except ScriptState.StateAssertException as _:
            pass
        return self

    # endregion

    # region Evaluation

    def script_evaluate(self, result):
        """
        This function should fill the Status and Message attributes of the result attribute
        based upon the results of the script using set_status and set_message

        Arguments:
            result (ScriptResult): The result

        Raises:
            NotImplementedError: There is not evaluation method defined, please define it with 'script_evaluate'
        """
        raise NotImplementedError("There is no evaluation method defined, please define it with 'script_evaluate'")

    def evaluate_script(self):
        """
        Evaluates the script result

        Returns:
            ScriptBase: The script
        """
        # Creates a new script result
        self._script_result = ScriptResult()
        self._script_result.set_status("green")
        self._script_result.set_message("")

        # If the state engine is being used
        if issubclass(self.__class__, ScriptState.StateEngine):
            try:
                # The script evaluate only load sets the state with the state engine
                self.script_evaluate(self._script_result)
            except NotImplementedError as _:
                pass
            finally:
                self: ScriptState.StateEngine
                state = self.get_state()
                self._script_result.set_status(state.status)
                self._script_result.set_message(state.message)
        else:
            self.script_evaluate(self._script_result)
        # Fills the Timestamp of the script result
        self._script_result.generate_timestamp()
        return self

    # endregion

    # region Output

    def set_output_path(self, path):
        """
        Sets the output path for the result of the script

        Arguments:
            path (str): The output path

        Raises:
            TypeError: The 'path' argument must be a str

        Returns:
            ScriptBase: The script
        """
        if type(path) != str:
            raise TypeError("The 'path' argument must be a str")
        self._script_output_path = path
        return self

    def export_to_output(self):
        """
        Exports the result of the script to a file path defined with set_output_path

        Raises:
            Exception: There is no output path specified

        Returns:
            ScriptBase: The script
        """
        if self._script_output_path is None:
            raise Exception("There is no output path specified")

        if self._script_result is None:
            self.evaluate_script()

        filePath = Path(self._script_output_path)
        filePath.parent.mkdir(parents=True, exist_ok=True)
        with open(filePath, "w") as file:
            file.write(self._script_result.generate_json())
        return self

    # endregion

    # region Result

    def log_output(self):
        """
        Logs the result of the script

        Returns:
            ScriptBase: The script
        """
        if self.get_internal_alias() is not None:
            print(f"--{self.get_internal_alias()}")
        print(f"Status: {self._script_result.Status}")
        print(f"Message: {self._script_result.Message}")
        return self

    def get_result(self):
        """
        Returns the ScriptResult of the script, it will evaluate the script if the script result doesn't exist

        Returns:
            ScriptResult: The script result of the script
        """
        if self._script_result is None:
            self.evaluate_script()
        return self._script_result

    # endregion

    # region Feeds

    def set_feed_guid(self, guid):
        """
        Sets the feed guid for the script

        Args:
            guid (str): The feed guid

        Returns:
            ScriptBase: The script
        """
        self._guid = guid

        return self

    def push_as_feed(self, access_token, secret):
        """
        Uses the script result as a feed and pushes the result

        Args:
            access_token (str): The access token of the api key
            secret (str): The secret of the api key
        """
        if self._guid is None:
            raise ValueError("No guid has been provided")

        feed = Feed()
        feed.set_script_result(self.get_result())
        feed.set_guid(self._guid)
        feed.set_api_key(access_token, secret)
        return feed.push()

    # endregion

    def get_internal_alias(self):
        """
        Returns an internal alias, if it is set, it can be used internally

        Returns:
            str: The internal alias
        """
        return self._script_alias

    def set_internal_alias(self, alias):
        """
        Sets a new internal alias

        Arguments:
            alias (str): The new internal alias

        Returns:
            ScriptBase: The class
        """
        self._script_alias = alias

        return self



from SeeThru_Feeds.Model.Attribution import Attribution
from SeeThru_Feeds.Model.Properties.Properties import *
from SeeThru_Feeds.Model.Properties.PropertyManager import PropertyManager


class ComponentBase(PropertyManager, Attribution):
    def component_execute(self):
        """
        This function should be overridden by a subclass
        This is where your component should start executing

        Raises:
            NotImplementedError: There is no execution method defined, please define it with 'component_execute'
        """
        raise NotImplementedError("There is no execution method defined, please define it with 'component_execute'")

    def run(self):
        """
        This method will call the sub class' component_execute method
        This is the only way a component should be executed as it 
        ensures that the properties of the component are valid

        Returns:
            ComponentBase: The component
        """
        self.check_fillables()
        # The fillable properties passed their parsing, therefore the component can be executed
        self.component_execute()
        return self



from datetime import datetime
from ..Exceptions.Config import HeaderException

class Header:
    Scheme_Name: str
    Scheme_Description: str
    Scheme_Author: str
    Scheme_Owner: str
    Creation_Date: str

    def __init__(self, name: str, description: str, author: str, owner: str, creation_date: str):
        self.Scheme_Name = name
        self.Scheme_Description = description
        self.Scheme_Author = author
        self.Scheme_Owner = owner
        self.Creation_Date = creation_date

    @staticmethod
    def new(name: str) -> "Header":
        """
        Creates a new config header, with default attributes

        Args:
            name (str): The name of the feedscheme

        Returns:
            ConfigHeader: The new header
        """
        description = "Enter a description for your feed scheme"
        author = "Enter the author of your feed scheme"
        owner = "Enter the owner for your feed scheme"
        creation_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        return Header(name, description, author, owner, creation_date)

    @staticmethod
    def load(data: dict) -> "Header":
        if "Scheme_Name" not in data:
            raise HeaderException("No Scheme_Name defined")
        if "Scheme_Description" not in data:
            raise HeaderException("No Scheme_Description defined")
        if "Scheme_Author" not in data:
            raise HeaderException("No Scheme_Author defined")
        if "Scheme_Owner" not in data:
            raise HeaderException("No Scheme_Owner defined")
        if "Creation_Date" not in data:
            raise HeaderException("No Creation_Data defined")
        return Header(
            data["Scheme_Name"],
            data["Scheme_Description"],
            data["Scheme_Author"],
            data["Scheme_Owner"],
            data["Creation_Date"]
        )

    def dump(self) -> dict:
        return {
            "Scheme_Name": self.Scheme_Name,
            "Scheme_Description": self.Scheme_Description,
            "Scheme_Author": self.Scheme_Author,
            "Scheme_Owner": self.Scheme_Owner,
            "Creation_Date": self.Creation_Date
        }


from unittest import TestCase


# TODO: as per TDD, all tests should fail initially, hence self.fail() below. But this is not a new repo, we will
#  hopefully add test coverage as we go along....
class TestSeeThru_Feed(TestCase):
    def test_create_feed_scheme(self):
        # self.fail()
        self.assertTrue(True)

    def test_create_script(self):
        # self.fail()
        self.assertTrue(True)

    def test_run_feed_scheme(self):
        # self.fail()
        self.assertTrue(True)

    def test_create_dir(self):
        # self.fail()
        self.assertTrue(True)

    def test_touch_file(self):
        # self.fail()
        self.assertTrue(True)




from SeeThru_Feeds.Model.Components.ComponentBase import ComponentBase
from SeeThru_Feeds.Model.Properties.Properties import FillableProperty, ResultProperty
import socket


class PortOpen(ComponentBase):
    TARGET_HOST = FillableProperty(
        name="target_host", required=True, of_type=str)
    PORT = FillableProperty(name="port", default=443,
                            required=True, of_type=int)
    SUCCEEDED = ResultProperty(name="succeeded")

    Component_Title = "PortOpen Socket Component"
    Component_Description = "This component will test to see if it can open a tcp connection with the given host and port"
    Component_Author = "SeeThru Networks"
    Component_Owner = "SeeThru Networks"

    def component_execute(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(2)

        try:
            # Attempts a connection to the socket
            sock.connect((self.get_property(PortOpen.TARGET_HOST),
                          self.get_property(PortOpen.PORT)))
            sock.close()
            self.set_property(PortOpen.SUCCEEDED, True)
        except:
            self.set_property(PortOpen.SUCCEEDED, False)


class UDPPortOpen(ComponentBase):
    TARGET_HOST = FillableProperty(
        name="target_host", required=True, of_type=str)
    PORT = FillableProperty(name="port", default=443,
                            required=True, of_type=int)
    # Stores whether the udp data send was successful
    SUCCEEDED = ResultProperty(name="succeeded")
    # If not successful, then this stores the os level error number for the socket connection
    ERROR_NO = ResultProperty(name="error_no", default=0)

    Component_Title = "UDPPortOpen Socket Component"
    Component_Description = "This component will test to see if it can send data over a udp socket with the given host and port"
    Component_Author = "SeeThru Networks"
    Component_Owner = "SeeThru Networks"

    def component_execute(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(5)

        try:
            # Attempts a connection to the socket
            sock = socket.socket(family=socket.AF_INET, type=socket.SOCK_DGRAM)
            sock.sendto(b"Test", (self.get_property(
                self.TARGET_HOST), self.get_property(self.PORT)))
            sock.close()
            self.set_property(self.SUCCEEDED, True)
        except socket.error as error:
            self.set_property(self.ERROR_NO, error.errno)
            self.set_property(self.SUCCEEDED, False)
        except:
            self.set_property(self.SUCCEEDED, False)



class InvalidPropertyType(Exception):
    def __init__(self, name, value, typeSet):
        self.name = name
        self.value = value
        self.typeSet = typeSet

    def __str__(self):
        return f"[Property: {self.name}] Type of {self.value} not in {self.typeSet}"


class InvalidPropertyValue(Exception):
    def __init__(self, name, value, valueSet):
        self.name = name
        self.value = value
        self.valueSet = valueSet

    def __str__(self):
        return f"[Property: {self.name}] {self.value} not in {self.valueSet}"


class PropertyValidatorError(Exception):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __str__(self):
        return f"[Property: {self.name}] {self.value} does not pass custom validator"


class PropertyRequired(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"[Property: {self.name}] Value not provided"



from copy import copy
from SeeThru_Feeds.Model.Properties.Exceptions import InvalidPropertyType, PropertyRequired, InvalidPropertyValue, \
    PropertyValidatorError


class PropertyBase:
    def __init__(self, name):
        if type(name) != str and name is not None:
            raise TypeError("Invalid type for 'name'")
        self.name = name
        self._value = None

    def parse_value(self, value):
        """
        Checks that the value given has met the conditions of the property

        Arguments:
            value (any): The value associated with this property

        Raises:
            NotImplementedError: The function hasn't been implemented yet for the subclass
            Exception: The property is None but it is required

        Returns:
            bool: Whether the value given is valid for the property
        """
        raise NotImplementedError("parse_value function not implemented yet for property")

    @staticmethod
    def new_instance(self):
        """Returns a new instance of the property, this is used by the property manager

        Returns:
            PropertyBase: The property copy
        """
        return copy(self)

    @property
    def value(self):
        """Provides a getter for the value of an instance of the property
        Parses the instance's value

        Returns:
            Any: The parsed value of the property
        """
        # The value should be provided by the PropertyManager
        if not hasattr(self, "_value"):
            raise Exception("Value attribute has not been set for property")

        return self._value

    @value.setter
    def value(self, value):
        """Sets the value to the property

        Args:
            value (any): The value to assign to the property

        """
        # Validates that the value given is valid
        if self.parse_value(value):
            self._value = value


class FillableProperty(PropertyBase):
    def __init__(self, name=None, default=None, required=False, of_type=None, value_set=None, custom=None):
        """
        Defines the fillable property with the parameters that the property should take

        Args:
            name (str): The internal name of the property
            default (any): A default value
            required (bool): Whether this property is required
            of_type (type, [type]): A type or a set of types that the property can be
            value_set (set, list, range]): A set of values that the property can be
            custom (lambda): A lambda function to check the value against
        """
        super().__init__(name)

        self.default = default

        if type(required) != bool:
            raise TypeError("Invalid type for 'required'")
        self.required = required

        self.ofType = of_type

        if type(value_set) == set or type(value_set) == list or type(value_set) == range:
            self.valueSet = set(value_set)
        elif value_set is None:
            self.valueSet = value_set
        else:
            raise TypeError("Invalid type for 'value_set'")

        if not callable(custom) and custom is not None:
            raise TypeError("Invalid type for 'func'")
        self.custom = custom

    def parse_value(self, value):
        """
        Ensures that the value provided is valid

        Args:
            value (any): The value to check against

        Returns:
            bool: Whether the value is valid

        """
        # --Required check
        if self.required and value is None:
            raise PropertyRequired(self.name)
        # --Type check, ensures that the value is of a type in of_type
        if self.ofType is not None:
            if type(self.ofType) == list:
                if type(value) not in self.ofType:
                    raise InvalidPropertyType(self.name, value, self.ofType)
            else:
                if value is not None and type(value) != self.ofType:
                    raise InvalidPropertyType(self.name, value, self.ofType)
        # --Value check, checks that the value is in the set given
        if self.valueSet is not None and value not in self.valueSet:
            raise InvalidPropertyValue(self.name, value, self.valueSet)
        # --Runs the custom function to check the value
        if self.custom is not None and not self.custom(value):
            raise PropertyValidatorError(self.name, value)
        return True


class ResultProperty(PropertyBase):
    def __init__(self, name=None, default=None, of_type=None, value_set=None, custom=None):
        """
        Defines the result property with the parameters that the property should take

        Args:
            name (str): The internal name of the property
            default (any): A default value
            of_type (type, [type]): A type or a set of types that the property can be
            value_set (set, list, range): A set of values that the property can be
            custom (lambda): A lambda function to check the value against
        """
        super().__init__(name)

        self.default = default

        self.ofType = of_type

        if type(value_set) == set or type(value_set) == list or type(value_set) == range:
            self.valueSet = set(value_set)
        elif value_set is None:
            self.valueSet = value_set
        else:
            raise TypeError("Invalid type for 'value_set'")

        if not callable(custom) and custom is not None:
            raise TypeError("Invalid type for 'func'")
        self.custom = custom

    def parse_value(self, value):
        # --Type check, ensures that the value is of a type in of_type
        if self.ofType is not None:
            if type(self.ofType) == list:
                if type(value) not in self.ofType:
                    raise InvalidPropertyType(self.name, value, self.ofType)
            else:
                if value is not None and type(value) != self.ofType:
                    raise InvalidPropertyType(self.name, value, self.ofType)
        # --Value check, checks that the value is in the set given
        if self.valueSet is not None and value not in self.valueSet:
            raise InvalidPropertyValue(self.name, value, self.valueSet)
        # --Runs the custom function to check the value
        if self.custom is not None and not self.custom(value):
            raise PropertyValidatorError(self.name, value)
        return True



import SeeThru_Feeds.Core.Exceptions.Config as ConfigExceptions


class Meta:
    Script_Name: str
    Script_Output_Path: str
    Script_Object_Path: str

    def __init__(self, name: str, output_path: str, object_path: str):
        self.Script_Name = name
        self.Script_Output_Path = output_path
        self.Script_Object_Path = object_path

    @staticmethod
    def new(name: str) -> "Meta":
        """
        Creates a new script meta, with attributes derived from the name

        Args:
            name (str): The name of the script

        Returns:
            ScriptMeta: The new script meta
        """
        output_path = f"Outputs/{name}"
        object_path = f"Scripts.{name}@{name}"
        return Meta(name, output_path, object_path)

    @staticmethod
    def load(data: dict, script_name: str) -> "Meta":
        """
        Parses the given data into a script meta

        Args:
            data: The data to parse
            script_name: The name of the script ot which this belongs

        Returns:
            Meta: The new script meta
        """
        if "Script_Name" not in data:
            raise ConfigExceptions.ScriptMetaException("A script meta has no name", script_name)
        if "Script_Output_Path" not in data:
            raise ConfigExceptions.ScriptMetaException("A script meta has no output path", script_name)
        if "Script_Object_Path" not in data:
            raise ConfigExceptions.ScriptMetaException("A script meta has no script module", script_name)
        return Meta(
            data["Script_Name"],
            data["Script_Output_Path"],
            data["Script_Object_Path"]
        )

    def dump(self) -> dict:
        """
        Dumps the meta into a dictionary

        Returns:
            dict: The dumped data
        """
        return {
            "Script_Name": self.Script_Name,
            "Script_Output_Path": self.Script_Output_Path,
            "Script_Object_Path": self.Script_Object_Path
        }



"""
Stores script state
A script can choose to use states rather than an evaluation method,
the script state gets set during the script runtime
Each state has a default status and message, these can be overwritten elsewhere if a custom status or message is needed
"""


class StateAssertException(Exception):
    pass


class State:
    """
    Defines a possible state, with a status and message pair
    """
    green = "green"
    ok = green
    amber = "amber"
    warning = amber
    red = "red"
    error = red

    def __init__(self, name, status, message):
        self._name = name
        self._status = status
        self._message = message

    @property
    def name(self):
        return self._name

    @property
    def status(self):
        return self._status

    @property
    def message(self):
        return self._message


class DefaultStates:
    """
    Defines the possible states that a script can have
    """
    OK = State("ok", State.ok, "")
    WARNING = State("warning", State.warning, "There is a warning")
    ERROR = State("error", State.error, "There is an error")


class StateEngine:
    """
    Stores evaluation states, this can be used as an alternative to script_evaluate,
    where the state is set as the script runs
    """
    class MessageStates(DefaultStates):
        """
        Creates the default state options, this can be overwritten if custom states are required
        """
        pass

    _state = DefaultStates.OK
    _configurations = {}

    def configure_state(self, state, status, message):
        """
        Configures a custom state for the script

        Args:
            state (str): The name of the state to configure
            status (str): The status of the state
            message (str): The message of the state

        Returns:
            bool: Whether the operation was succesfull
        """
        status = status.lower()
        if status not in ["green", "amber", "red"]:
            raise ValueError("Please provide a valid status")
        # Matches the state with an existing state
        for variable in dir(self.MessageStates):
            var = getattr(self.MessageStates, variable)
            if isinstance(var, State):
                if var.name == state:
                    # Stores the state in the configuration, not overriding the MessageStates
                    self._configurations[state] = State(state, status, message)
                    # Checks if the currently set state should be reset
                    if self._state.name == state:
                        self._state = self._configurations[state]
                    return True
        return False

    def set_state(self, state):
        """
        Sets the state of the script
        This can be used as an alternative to evaluation

        Args:
            state (State): The state

        Returns:
            ScriptBase: The script
        """
        if type(state) != State:
            raise TypeError("Please provide a valid state")
        # Uses the state in the configurations if one exists for the state name
        if state.name in self._configurations:
            self._state = self._configurations[state.name]
        else:
            self._state = state

        return self

    def get_state(self):
        """
        Returns the state of the script

        Returns:
           State: The state of the script
        """
        return self._state

    def assert_true(self, condition, state):
        """
        Asserts that the condition is true,
        If not, the state is set and the script is ended

        Args:
            condition (bool): The condition
            state (State): The state to set to
        """
        if not condition:
            self._state = state
            # Raises an exception to escape the run method
            raise StateAssertException()



from datetime import datetime


class ScriptResult:
    def __init__(self, status=None, message=None):
        """
        Creates a new script result

        Args:
            status: The status of the result
            message: The message of the result
        """
        self.Status = status
        self.Message = message
        self.Timestamp = None

    def set_status(self, status):
        """
        Sets the Status of the result to a specific value

        Arguments:
            status (str): The status of the result

        Returns:
            ScriptResult: The script result

        Raises:
            TypeError: Invalid type of 'Status', it must be str
            ValueError: Invalid value for 'Status'
        """
        if type(status) != str:
            raise TypeError("Invalid type of 'Status', it must be str")
        if status.lower() not in ["red", "amber", "green"]:
            raise ValueError("Invalid value for 'Status'")
        self.Status = status.lower()

        return self

    def set_message(self, message):
        """
        Sets the Message of the result to a specific value

        Arguments:
            message (str): The message of the result

        Returns:
            ScriptResult: The script result

        Raises:
            TypeError: Invalid type of 'Message', it must be str
            ValueError: Message is too long, it must be below 256 characters
        """
        if type(message) != str:
            raise TypeError("Invalid type of 'Message', it must be str")
        self.Message = message

        return self

    def generate_timestamp(self):
        """
        Generates a Timestamp for the result

        Returns:
            ScriptResult: The script result
        """
        self.Timestamp = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")

        return self

    def generate_json(self):
        """
        Generates the json result result

        Returns:
            str: The json string
        """
        if self.Status is None or self.Message is None or self.Timestamp is None:
            raise Exception("Not all values are provided for the result")
        return "{\"color\":\"%s\", \"Message\":\"%s\", \"time\":\"%s\"}" % (self.Status, self.Message, self.Timestamp)

    def get_status(self):
        """
        Returns the Status of the result

        Returns:
            str: The Status
        """
        return self.Status

    def get_message(self):
        """
        Returns the Message of the result

        Returns:
            str: The Message
        """
        return self.Message

    def get_timestamp(self):
        """
        Returns the status of the script result

        Returns:
            str: The timestamp
        """
        return self.Timestamp



from SeeThru_Feeds.Model.Scripts.ScriptState import State, DefaultStates, StateEngine
from SeeThru_Feeds.Model.Scripts.ScriptBase import ScriptBase
from SeeThru_Feeds.Model.Properties.Properties import FillableProperty, ResultProperty
from SeeThru_Feeds.Library.Components.Socket import PortOpen


class TCPPortOpen(ScriptBase, StateEngine):
    class MessageStates(DefaultStates):
        port_closed = State("port_closed", State.error, "Cannot create a tcp socket to given host and port")

    host = FillableProperty(name="host", required=True, of_type=str)
    port = FillableProperty(name="port", required=False, of_type=[str, int], default=443)

    is_port_open = ResultProperty(name="is_port_open")

    Attr_Title = "TCP Port Open"
    Attr_Description = "A script which tests creating a tcp socket to a given host and port"
    Attr_Author = "SeeThru Networks"
    Attr_Owner = "SeeThru Networks"

    # ------ Script Overrides ------
    def script_run(self):
        # Checks to see if the port is open for tcp connections
        is_port_open = PortOpen()\
            .set_property(PortOpen.TARGET_HOST, self.host.value)\
            .set_property(PortOpen.PORT, self.port.value)\
            .run()\
            .get_property(PortOpen.SUCCEEDED)

        self.is_port_open.value = is_port_open

        # Sets the state to port closed
        if not self.is_port_open.value:
            self.set_state(self.MessageStates.port_closed)



import SeeThru_Feeds.Core.Exceptions.Config as ConfigExceptions

class State:
    Name: str
    Status: str
    Message: str

    def __init__(self, name: str, status: str, message: str):
        self.Name = name
        self.Status = status
        self.Message = message

    @staticmethod
    def new(name: str, status: str, message: str) -> "State":
        """
                Creates a new script state, with given attributes

                Args:
                    name (str): The name of the state
                    status (str): The status to give the script state
                    message (str): The message to give the script state

                Returns:
                    State: The new script state

                """
        return State(name, status, message)

    @staticmethod
    def load(data: dict, script_name: str) -> "State":
        """
        Parses the given data into a script state
        Args:
            data: The data to parse
            script_name: The name of the script

        Returns:
            State: The state
        """
        if "Name" not in data:
            raise ConfigExceptions.ScriptStateException("A script state has no name", script_name)
        if "Status" not in data:
            raise ConfigExceptions.ScriptStateException(f"The script state {data['Name']} has no status", script_name)
        if "Message" not in data:
            raise ConfigExceptions.ScriptStateException(f"The script state {data['Name']} has no message", script_name)
        return State(
            data["Name"],
            data["Status"],
            data["Message"]
        )

    def dump(self) -> dict:
        """
        Dumps the state into a dictionary

        Returns:
            dict: The dumped data
        """
        return {
            "Name": self.Name,
            "Status": self.Status,
            "Message": self.Message
        }



class Attribution:
    # region Title
    # The Script_Title attribute should be set in your implementation
    Attr_Title = None

    @classmethod
    def get_title(cls):
        """
        Returns a set title

        Raises:
            NotImplementedError: There is no title defined, please define it as 'Attr_Title='

        Returns:
            str: The title
        """
        if cls.Attr_Title is None:
            raise NotImplementedError("There is no title defined, please define it as 'Attr_Title='")
        return cls.Attr_Title
    # endregion

    # region Description
    # The Attr_Description attribute should be set in your implementation
    Attr_Description = None

    @classmethod
    def get_description(cls):
        """
        Returns a description

        Raises:
            NotImplementedError: There is no description defined, please define it as 'Attr_Description='

        Returns:
            str: The description
        """
        if cls.Attr_Description is None:
            raise NotImplementedError("There is no description defined, please define it as 'Attr_Description='")
        return cls.Attr_Description
    # endregion

    # region Author
    # The Attr_Author attribute should be set in your implementation
    Attr_Author = None

    @classmethod
    def get_author(cls):
        """
        Returns the author

        Raises:
            NotImplementedError: There is no author defined, please define it as 'Attr_Author='

        Returns:
            str: The author
        """
        if cls.Attr_Author is None:
            raise NotImplementedError("There is no author defined, please define it as 'Attr_Author='")
        return cls.Attr_Author
    # endregion

    # region Owner
    # The Attr_Owner attribute should be set in your implementation
    Attr_Owner = None

    @classmethod
    def get_owner(cls):
        """
        Returns a owner

        Raises:
            NotImplementedError: There is no owner defined, please define it as 'Attr_Owner='

        Returns:
            str: The owner
        """
        if cls.Attr_Owner is None:
            raise NotImplementedError("There is no owner defined, please define it as 'Attr_Owner='")
        return cls.Attr_Owner
    # endregion

    # region Support_Link
    # The Attr_SupportLink attribute should be set in your implementation
    Attr_SupportLink = None

    @classmethod
    def get_support_link(cls):
        """
        Returns a support link

        Raises:
            NotImplementedError: There is no supportLink defined, please define it as 'Attr_SupportLink='

        Returns:
            str: The support link
        """
        if cls.Attr_SupportLink is None:
            raise NotImplementedError("There is no supportLink defined, please define it as 'Attr_SupportLink='")
        return cls.Attr_SupportLink
    # endregion

    # region Docs_Link
    # The Attr_DocLink attribute should be set in your implementation
    Attr_DocLink = None

    @classmethod
    def get_docs_link(cls):
        """
        Returns a doc link

        Raises:
            NotImplementedError: There is no DocLink defined, please define it as 'Attr_DocLink='

        Returns:
            str: The doc link
        """
        if cls.Attr_DocLink is None:
            raise NotImplementedError("There is no DocLink defined, please define it as 'Attr_DocLink='")
        return cls.Attr_DocLink
    # endregion

    # region LicenseLink
    # The Attr_LicenseLink attribute should be set in your implementation
    Attr_LicenseLink = None

    @classmethod
    def get_license_link(cls):
        """
        Returns a license link

        Raises:
            NotImplementedError: There is no LicenseLink defined, please define it as 'Attr_LicenseLink='

        Returns:
            str: The license link
        """
        if cls.Attr_LicenseLink is None:
            raise NotImplementedError("There is no LicenseLink defined, please define it as 'Attr_LicenseLink='")
        return cls.Attr_LicenseLink
    # endregion



class ConfigException(Exception):
    pass


class NoConfigException(ConfigException):
    pass


class HeaderException(ConfigException):
    pass


class ScriptException(ConfigException):
    script_name: str

    def __init__(self, message, script_name):
        super().__init__(message)

        self.script_name = script_name


class ScriptMetaException(ScriptException):
    pass


class ScriptStateException(ScriptException):
    pass


class ScriptFillableException(ScriptException):
    name: str

    def __init__(self, message, script_name, fillable_name):
        super().__init__(message, script_name)

        self.name = fillable_name


class FeedException(ConfigException):
    name: str

    def __init__(self, message, feed_name):
        super().__init__(message)

        self.name = feed_name


class ApiKeyException(ConfigException):
    name: str

    def __init__(self, message, api_key_name):
        super().__init__(message)

        self.name = api_key_name


if __name__ == "__main__":
    raise ScriptStateException("", "", "")



import json
import sys
from datetime import datetime
from pathlib import Path
from typing import Dict, Any

from .Exceptions import Config as ConfigExceptions
from .Config import Config

import toml


class ConfigParser:
    def __init__(self, config_path: Path, method: str = "json"):
        self.config_path: Path = config_path
        self.method: str = method
        self.config: Config = Config.new("")

    @staticmethod
    def json(config_path: Path):
        """
        Creates and sets the parser's type to json

        Args:
            config_path (Path):The config file path

        Returns:
            ConfigParser: The parser
        """
        return ConfigParser(config_path, "json")

    @staticmethod
    def toml(config_path: Path):
        """
        Creates and sets the parser's type to toml

        Args:
            config_path (Path):The config file path

        Returns:
            ConfigParser: The parser
        """
        return ConfigParser(config_path, "toml")

    def set_config(self, config: Config):
        """
        Sets the config to the parser

        Args:
            config (Config): The config

        Returns:
            ConfigParser: The parser
        """
        self.config = config
        return self

    def load(self):
        """
        Opens the config file and parses it into the object model

        Returns:
            Config: The config
        """
        # Opens the config file and parses it
        if not self.config_path.exists():
            return self.config
        with self.config_path.open('r') as file:
            data = {}
            if self.method == "json":
                # Generates the config
                data = json.loads(file.read())
            elif self.method == "toml":
                # Generates the config
                data = toml.loads(file.read())

        try:
            self.config = Config.load(data)
            return self.config
        except ConfigExceptions.ScriptException as e:
            print(f"{e.script_name}: {e}")
            sys.exit(1)
        except ConfigExceptions.FeedException as e:
            print(f"{e.name}: {e}")
            sys.exit(1)
        except ConfigExceptions.ApiKeyException as e:
            print(f"{e.name}: {e}")
            sys.exit(1)
        except ConfigExceptions.ConfigException as e:
            print(e)
            sys.exit(1)

    def save(self):
        """
        Saves the config file, with the appropriate type
        """
        if not self.config_path.parent.exists():
            return
        with self.config_path.open('w') as file:
            content = ""
            if self.method == "json":
                content = json.dumps(self.config.dump(), indent=4)
            elif self.method == "toml":
                content = toml.dumps(self.config.dump())
            file.write(content)

    def __enter__(self):
        return self.load()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()


if __name__ == "__main__":
    # config = Config.new("Generic")
    # script = config.add_script("Google")
    # script.add_fillable("HOST", "192.168.0.1")
    #
    # output = config.dump()
    # print(output)

    with ConfigParser.toml(Path("./config.toml")) as config:
        breakpoint()
